#include "CGame.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image/stb_image.h"
#include "Model.h"
#include "Mesh.h"
#include "glm/mat4x4.hpp"
#include "glm/vec3.hpp"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtc/type_ptr.hpp"
#include <chrono>

void CGame::Initialize(unsigned int screenWidth, unsigned int screenHeight)
{
    m_Width = screenWidth;
    m_Height = screenHeight;

    SetupGPUProgram();
    SetupModel();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_MULTISAMPLE);
    glEnable(GL_STENCIL_TEST);

    GLuint depthStencilId;

    glGenFramebuffers(1, &m_IntermediateFBOId);
    glBindFramebuffer(GL_FRAMEBUFFER, m_IntermediateFBOId);

    glGenTextures(1, &m_ColorTextureId);
    glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, m_ColorTextureId);
    glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, 4, GL_RGB8, m_Width, m_Height, GL_TRUE);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D_MULTISAMPLE, m_ColorTextureId, 0);

    glGenTextures(1, &m_DepthStencilTextureId);
    glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, m_DepthStencilTextureId);
    glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, 4, GL_DEPTH24_STENCIL8, screenWidth, screenHeight, GL_TRUE);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D_MULTISAMPLE, m_DepthStencilTextureId, 0);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) == GL_FRAMEBUFFER_COMPLETE)
    {
        int a = 0;
        ++a;
    }

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void CGame::Shutdown()
{
    
}

void CGame::OnResize(int width, int height)
{
    m_Width = width;
    m_Height = height;

    glViewport(0, 0, m_Width, m_Height);
    glScissor(0, 0, m_Width, m_Height);
}

void CGame::Render()
{
    static auto startTime = std::chrono::high_resolution_clock::now();

    auto currentTime = std::chrono::high_resolution_clock::now();
    float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

    glEnable(GL_MULTISAMPLE);
    glBindFramebuffer(GL_FRAMEBUFFER, m_IntermediateFBOId);

    glEnable(GL_STENCIL_TEST);
    glClearColor(0.2f, 0.2f, 0.2f, 1.0f);
    glClearDepth(1.0f);
    glClearStencil(0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // Draw the Filled Man
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    // Fill the stencil buffer with 1 anywhere our tekka model is drawn.
    float scaleFactor = 0.4f;
    glm::mat4 model = glm::scale(glm::mat4(1.0f), glm::vec3(scaleFactor, scaleFactor, scaleFactor));
    model = glm::rotate(model, glm::radians(45.0f * time), glm::vec3(0, 1, 0));
    glm::mat4 view;
    glm::vec3 eyePos = glm::vec3(0.0f, -1.0f, -2.5f);
    view = glm::translate(glm::mat4(1.0f), eyePos);
    glm::mat4 projection;
    projection = glm::perspective(glm::radians(60.0f), m_Width / static_cast<float>(m_Height), 0.03f, 100.0f);
    glm::mat4 mvp = projection * view * model;
    glm::vec3 diffuseColor = glm::vec3(1,0,0);

    m_pShaderPass0->use();
    m_pShaderPass0->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass0->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass0->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass0->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass0->setVector3("eyePos", glm::value_ptr(eyePos));
    m_pShaderPass0->setVector3("diffuseColor", glm::value_ptr(diffuseColor));
    m_pShaderPass0->setFloat("time", time);
    m_pManModel->meshes[0].Draw(*m_pShaderPass0);

    // Additive Point Light Pass
    glm::vec3 pointLightColor = glm::vec3(0, 1, 0);
    glm::vec3 pointLightPositionInViewSpace = glm::vec3(2, 0, 5);

    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_ONE, GL_ONE);
    m_pShaderPass1->use();
    m_pShaderPass1->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass1->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass1->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass1->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass1->setVector3("eyePos", glm::value_ptr(eyePos));
    m_pShaderPass1->setVector3("pointLightPosition", glm::value_ptr(pointLightPositionInViewSpace));
    m_pShaderPass1->setVector3("pointLightColor", glm::value_ptr(pointLightColor));
    m_pShaderPass1->setFloat("time", time);
    m_pManModel->meshes[0].Draw(*m_pShaderPass1);
    glDisable(GL_BLEND);

    // Additive Point Light Pass
    pointLightColor = glm::vec3(0, 0, 1);
    pointLightPositionInViewSpace = glm::vec3(0, 2, 2);

    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_ONE, GL_ONE);
    m_pShaderPass1->use();
    m_pShaderPass1->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass1->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass1->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass1->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass1->setVector3("eyePos", glm::value_ptr(eyePos));
    m_pShaderPass1->setVector3("pointLightPosition", glm::value_ptr(pointLightPositionInViewSpace));
    m_pShaderPass1->setVector3("pointLightColor", glm::value_ptr(pointLightColor));
    m_pShaderPass1->setFloat("time", time);
    m_pManModel->meshes[0].Draw(*m_pShaderPass1);
    glDisable(GL_BLEND);

    // Draw a Floor quad
    diffuseColor = glm::vec3(1, 1, 1);
    model = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -0.2f, 0.0f));
    model = glm::scale(model, glm::vec3(5.0f, 5.0f, 5.0f));
    model = glm::rotate(model, glm::radians(45.0f * time), glm::vec3(0, 1, 0));
    mvp = projection * view * model;

    m_pShaderPass0->use();
    m_pShaderPass0->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass0->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass0->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass0->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass0->setVector3("eyePos", glm::value_ptr(eyePos));
    m_pShaderPass0->setFloat("time", time);
    m_pShaderPass0->setVector3("diffuseColor", glm::value_ptr(diffuseColor));
    m_pQuad->meshes[0].Draw(*m_pShaderPass0);

    // Blit intermediate FBO to backbuffer FBO

    glBindFramebuffer(GL_READ_FRAMEBUFFER, m_IntermediateFBOId);
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
    glBlitFramebuffer(0, 0, m_Width, m_Height, 0, 0, m_Width, m_Height, GL_COLOR_BUFFER_BIT, GL_NEAREST);
}

void CGame::SetupModel()
{
    m_pManModel = new Model();
    m_pManModel->Load("man.obj");

    m_pQuad = new Model();
    m_pQuad->Load("quad.obj");
}

void CGame::SetupGPUProgram()
{
    m_pShaderPass0 = new Shader("vertexPass0.glsl", "fragmentPass0.glsl");
    m_pShaderPass1 = new Shader("vertexPass1.glsl", "fragmentPass1.glsl");
}

void CGame::ReloadShaders()
{
    OutputDebugStringA("Reloading Shaders!\n");
    SetupGPUProgram();
}